/* colong.c: calculate lunar colongitude.
 * Copyright 2001 by Akkana Peck, akkana <at> shallowsky.com.
 * You are free to free to use or modify this software under the GPL.
 *
 * Tables and algorithm from Rukl p. 212.
 */

#include "feature.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define Dprintf if (0) printf

extern double GetTimeZoneOffset();
double TimeZone = -7;   // default to PDT

typedef struct coLongCorrec_s {
  double c0, M0;
} coLongCorrect;

/* Table I: Year.  Starts at 1989. */
static coLongCorrect yearCorrec[] =
{
    { 179.32, 0.9 },
    { 308.95, 0.7 },    /* 1990 */
    {  78.57, 0.4 },
    { 220.39, 1.1 },
    { 350.01, 0.9 },
    { 119.63, 0.6 },
    { 249.26, 0.4 },    /* 1995 */
    {  31.07, 1.1 },
    { 160.69, 0.8 },
    { 290.32, 0.6 },
    {  59.94, 0.3 },
    { 201.76, 1.0 },    /* 2000 */
    { 331.38, 0.7 },
    { 101.01, 0.5 },
    { 230.63, 0.2 },
    {  12.44, 1.0 },
    { 142.07, 0.7 },    /* 2005 */
    { 271.69, 0.5 },
    {  41.31, 0.2 },
    { 183.13, 0.9 },
    { 312.75, 0.7 },
    {  82.38, 0.4 },    /* 2010 */
    { 212.00, 0.2 },
    { 353.81, 0.9 },
    { 123.44, 0.6 },
    { 253.06, 0.4 },
    {  22.68, 0.1 },    /* 2015 */
    { 164.50, 0.8 },
    { 294.12, 0.6 },
    {  63.75, 0.3 },
    { 193.37, 0.1 },
    { 335.18, 0.8 },    /* 2020 */
};

static coLongCorrect monthCorrec[] =
{
    {   0.00,   0.0 },
    {   0.00, 356.0 },
    {  17.91,  26.6 },
    { 359.25,  54.2 },
    {  17.17,  84.7 },
    {  22.89, 114.3 },
    {  40.80, 144.8 },
    {  46.53, 174.4 },
    {  64.44, 204.9 },
    {  82.35, 235.5 },
    {  88.07, 265.1 },
    { 105.99, 295.6 },
    { 111.71, 325.2 },
    { 347.81, 355.0 },    /* leap year Januaries */
    {   5.72,  25.6 },    /* leap year Februaries */
};

static coLongCorrect dayCorrec[] =
{
    {   0.00, 0.0 },
    {  12.19, 1.0 },
    {  24.38, 2.0 },
    {  36.57, 3.0 },
    {  48.76, 3.9 },
    {  60.95, 4.9 },
    {  73.14, 5.9 },
    {  85.34, 6.9 },
    {  97.53, 7.9 },
    { 109.72, 8.9 },
    { 121.91, 9.9 },
    { 134.10, 10.8 },
    { 146.29, 11.8 },
    { 158.48, 12.8 },
    { 170.67, 13.8 },
    { 182.86, 14.8 },
    { 195.05, 15.8 },
    { 207.24, 16.8 },
    { 219.43, 17.7 },
    { 231.62, 18.7 },
    { 243.81, 19.7 },
    { 256.01, 20.7 },
    { 268.20, 21.7 },
    { 280.39, 22.7 },
    { 292.58, 23.7 },
    { 304.77, 24.6 },
    { 316.96, 25.6 },
    { 329.15, 26.6 },
    { 341.34, 27.6 },
    { 353.53, 28.6 },
    {   5.72, 29.6 },
    {  17.91, 30.6 },
};

static coLongCorrect hourCorrec[] =
{
    {  0.00, 0.0 },
    {  0.51, 0.0 },
    {  1.02, 0.1 },
    {  1.52, 0.1 },
    {  2.03, 0.2 },
    {  2.54, 0.2 },
    {  3.05, 0.2 },
    {  3.56, 0.3 },
    {  4.06, 0.3 },
    {  4.57, 0.4 },
    {  5.08, 0.4 },
    {  5.59, 0.5 },
    {  6.10, 0.5 },
    {  6.60, 0.5 },
    {  7.11, 0.6 },
    {  7.62, 0.6 },
    {  8.13, 0.7 },
    {  8.64, 0.7 },
    {  9.14, 0.7 },
    {  9.65, 0.8 },
    { 10.16, 0.8 },
    { 10.67, 0.9 },
    { 11.17, 0.9 },
    { 11.68, 0.9 },
    { 12.19, 1.0 },
};

static double finalCorrec[] = {
    0.0, 0.4, 0.7, 1.0, 1.3, 1.5, 1.7, 1.8, 1.9, 1.9, 
    1.9, 1.8, 1.6, 1.5, 1.2, 1.0, 0.6, 0.3, 0.0
};

static int isLeapYear(int year)
{
    if (year%4 == 0) return 1;
    if (year%100 == 0 && year%1000 != 0) return 1;
    return 0;
}

/* For DaysInMonth, month goes 1-12 for this fcn;
 * year is the complete (4-digit) year.
 */
static int DaysInMonth(int m, int y)
{
    if (m == 2)
        return isLeapYear(y) ? 29 : 28;
    if (m == 4 || m == 6 || m == 9 || m == 11)
        return 30;
    return 31;
}

/* Get the colongitude, from time parameters in UT */
static double colongitude(int year, int month, int day, int hour, int minute)
{
    double c, M;
    c = yearCorrec[year - 1989].c0;
    M = yearCorrec[year - 1989].M0;

    if (isLeapYear(year))
        if (month < 2) month += 12;
    c += monthCorrec[month].c0;
    M += monthCorrec[month].M0;

    c += dayCorrec[day].c0;
    M += dayCorrec[day].M0;

    c += hourCorrec[hour].c0;
    M += hourCorrec[hour].M0;

    /* Now apply the correction factor to c based on M */
    if (M > 180)
        c += finalCorrec[36 - (int)(M/10)];
    else
        c -= finalCorrec[(int)(M/10)];

    /* Rukl doesn't have an equation for minutes,
     * but even a rough average is better than nothing.
     */
    c += (minute / 60.) * (hourCorrec[hour+1].c0 - hourCorrec[hour].c0);

    c = fmod(c, 360.L);

    return c;
}

/* When will a particular colongitude happen in a given month?
 * Return the day and hour it will happen.
 * The struct tm passed in is the starting point,
 * and we will search forward from there.
 */
static void
timeOfEvent(struct tm *then, double *targetLong)
{
    //double c, M;
    int d;
    int h;
    double co;
    int needWrap = 0;
    int first;
    int year = then->tm_year + 1900;
    int month = then->tm_mon+1;

    Dprintf("Calculating colong %lf in month %d of %d starting from %d\n",
            *targetLong, month, year, then->tm_mday);
    //c = yearCorrec[year-1989].c0 + monthCorrec[month].c0;
    //M = yearCorrec[year-89].M0 + monthCorrec[month].M0;

    /* Find the day after this happens. */
    co = colongitude(then->tm_year+1900, month, then->tm_mday, 0, 0);
    // If it's already bigger, than we have to keep iterating until
    // the colongitude wraps around past 360 degrees.
    if (co > *targetLong) {
        needWrap = 1;
        first = co;
    }
    for (d = then->tm_mday; d <= DaysInMonth(month, year); ++d)
    {
        co = colongitude(year, month, d, 0, 0);
        if (needWrap)
        {
            if (co > first)
                continue;
            needWrap = 0;
        }
        if (co == *targetLong)
        {
            then->tm_mday = d;
            then->tm_hour = 0;
            then->tm_min = 0;
            then->tm_sec = 0;
            *targetLong = co;
            return;
        }
        if (co > *targetLong)
            break;
        Dprintf("Day loop: %d/%d/%d 0h: %d\n", month, d, year, (int)co);
    }
    --d;       /* go back to the previous day */

    /* Now loop forward looking for the right time. */
    if (co > *targetLong) {
        needWrap = 1;
        first = co;
    }
    while (d < DaysInMonth(month, year))
    {
        for (h = 1; h <= 24; ++h)
        {
            co = colongitude(year, month, d, h, 0);
            if (needWrap)
            {
                if (co > first)
                    continue;
                needWrap = 0;
            }
            Dprintf("Hour loop: %d/%d/%d %dh: %.2lf\n", month, d, year, h, co);
                   
            if (co > *targetLong)
            {
                /* We actually want to give prior warning,
                 * so back up to the previous hour:
                 */
                --h;
                *targetLong = colongitude(year, month, d, h, 0);
                then->tm_mday = d;
                then->tm_hour = h;
                Dprintf("Returning from hour loop, c = %.2lf\n", *targetLong);
                return;
            }
        }
        ++d;
    }
    /* If we get here, then we didn't find it in this month.
     * Loop forward to the next one.
     */
    ++(then->tm_mon);
    then->tm_mday = 1;
    timeOfEvent(then, targetLong);
    //printf("PANIC!  Never found it.\n");
}

static int parseMonth(char* mon)
{
    if (!strncasecmp(mon, "jan", 3))
        return 0;
    else if (!strncasecmp(mon, "feb", 3))
        return 1;
    else if (!strncasecmp(mon, "mar", 3))
        return 2;
    else if (!strncasecmp(mon, "apr", 3))
        return 3;
    else if (!strncasecmp(mon, "may", 3))
        return 4;
    else if (!strncasecmp(mon, "jun", 3))
        return 5;
    else if (!strncasecmp(mon, "jul", 3))
        return 6;
    else if (!strncasecmp(mon, "aug", 3))
        return 7;
    else if (!strncasecmp(mon, "sep", 3))
        return 8;
    else if (!strncasecmp(mon, "oct", 3))
        return 9;
    else if (!strncasecmp(mon, "nov", 3))
        return 10;
    else if (!strncasecmp(mon, "cec", 3))
        return 11;
    return -1;
}

static void Usage(char** argv)
{
    printf("Usage: %s [month day [year] hh:mm]\n", argv[0]);
    printf("          Calculate colongitude for the given date.\n");
    printf("       %s -t target_long [month [year]]\n", argv[0]);
    printf("          Find time in the given month when target colong will be seen.\n");
    printf("       %s -f featurename [month [year]]\n", argv[0]);
    printf("          Find time in the given month when target feature will be seen.\n");
    printf("Dates default to current time.\n");
    exit(1);
}

int main(int argc, char **argv)
{
    struct tm then, *thenp;
    time_t sec;
    int argptr;
    int targetMode = 0;
    int featureMode = 0;
    int briefMode = 0;       /* Used when we're called from a CGI */
    double targetLong;
    struct LunarFeature* feature = 0;

    --argc;     // get rid of progname

    // First, extract flags:
    for (argptr = 1; argc > 0; )
    {
        if (argv[argptr][0] != '-')
            break;

        if (argv[argptr][1] == 't')
            targetMode = 1;
        else if (argv[argptr][1] == 'f')
            featureMode = 1;
        else if (argv[argptr][1] == 'b')
            briefMode = 1;
        else Usage(argv);
        ++argptr;
        --argc;
    }
    if (targetMode && featureMode)
        Usage(argv);

    if (targetMode)
    {
        if (argc < 1)
            Usage(argv);
        targetLong = atof(argv[argptr++]);
        --argc;
    }

    else if (featureMode)
    {
        if (argc < 1)
            Usage(argv);
        feature = GetFeatureByKeyword(argv[argptr], "features");
        if (!feature) {
            printf("Couldn't find feature %s\n", argv[argptr]);
            return 1;
        }
        targetMode = 1;
        targetLong = feature->colong;
        ++argptr;
        --argc;
    }

    Dprintf("Down to date, argc = %d\n", argc);
    // Fill in "then" with current date; we'll override specifics
    time(&sec);
    thenp = gmtime(&sec);       // consider using localtime() ?
    then = *thenp;              // don't trust libs not to overwrite

    if (argc > 0)
    {
        Dprintf("Getting month from '%s'\n", argv[argptr]);
        if (isdigit(argv[argptr][0]))
            then.tm_mon = atoi(argv[argptr++]) - 1;
        else
            then.tm_mon = parseMonth(argv[argptr++]);
        if (then.tm_mon < 0) {
            printf("Illegal month: %s\n", argv[argptr-1]);
            return 1;
        }
        --argc;
    }

    if (targetMode)     // remaining arguments: [year]
    {
        if (argc > 1) Usage(argv);
        if (argc == 1)
            then.tm_year = atoi(argv[argptr++]);
        then.tm_hour = 0;
        then.tm_min = 0;
    }
    else                // remaining arguments: day [year] hh:mm 
    {
        if (argc > 3) Usage(argv);
        if (argc > 0)
        {
            then.tm_mday = atoi(argv[argptr++]);
            --argc;
        }
        if (argc > 1)
        {
            then.tm_year = atoi(argv[argptr++]);
            --argc;
        }
        if (argc > 0)
        {
            char* hour = argv[argptr++];
            char* colon = strchr(hour, ':');
            if (colon > hour) {
                char* min = colon;
                *(min++) = '\0';
                then.tm_min = atoi(min);
            }
            else then.tm_min = 0;
            then.tm_hour = atoi(hour);
            --argc;
        }
    }

    /* A struct tm year is actually year - 1900, hah.
     * Try to guess what the user had in mind.
     */
    if (then.tm_year > 200)
        then.tm_year -= 1900;
    else if (then.tm_year < 25)
        then.tm_year += 100;

    if (targetMode)
    {
        timeOfEvent(&then, &targetLong);
    }
    else
    {
        targetLong = colongitude(then.tm_year+1900, then.tm_mon+1,
                                 then.tm_mday,
                                 then.tm_hour, then.tm_min);
    }
    if (featureMode && feature)
        printf("%s: ", feature->longdesc);
    if (briefMode && targetMode)
        printf("%02d/%02d/%02d %02d:%02d\n",
               then.tm_mon+1, then.tm_mday, then.tm_year+1900,
               then.tm_hour, then.tm_min);
    else
        printf("%.2lf on %02d/%02d/%04d UT %02d:%02d\n",
               targetLong, then.tm_mon+1, then.tm_mday, then.tm_year+1900,
               then.tm_hour, then.tm_min);

    return 0;
}

